# Copyright (c) 2013, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of Intel Corporation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Give a presentation using EFI Graphics Output Protocol (GOP)."""

import bits
import efi
import itertools
import os
import readline
import struct
import unpack

PixelRedGreenBlueReserved8BitPerColor, PixelBlueGreenRedReserved8BitPerColor, PixelBitMask, PixelBltOnly = range(4)

EfiBltVideoFill, EfiBltVideoToBltBuffer, EfiBltBufferToVideo, EfiBltVideoToVideo = range(4)

class GraphicsOutputModeInformation(unpack.Struct):
    """Decode an EFI Graphics Output Mode Information structure"""
    def __init__(self, gop_mode):
        super(GraphicsOutputModeInformation, self).__init__()
        u = unpack.Unpackable(bits.memory(gop_mode.InfoPtr, gop_mode.SizeOfInfo))
        self.add_field("Version", u.unpack_one("I"))
        self.add_field("HorizontalResolution", u.unpack_one("I"))
        self.add_field("VerticalResolution", u.unpack_one("I"))
        self.add_field("PixelFormat", u.unpack_one("I"))
        self.add_field("RedMask", u.unpack_one("I"))
        self.add_field("GreenMask", u.unpack_one("I"))
        self.add_field("BlueMask", u.unpack_one("I"))
        self.add_field("ReservedMask", u.unpack_one("I"))
        self.add_field("PixelsPerScanLine", u.unpack_one("I"))

class GraphicsOutputProtocolMode(unpack.Struct):
    """Decode an EFI Graphics Output Protocol Mode"""
    def __init__(self, ptr):
        super(GraphicsOutputProtocolMode, self).__init__()
        u = unpack.Unpackable(bits.memory(ptr, struct.calcsize("IIPPQP")))
        self.add_field("MaxMode", u.unpack_one("I"))
        self.add_field("Mode", u.unpack_one("I"))
        self.add_field("InfoPtr", u.unpack_one("P"))
        self.add_field("SizeOfInfo", u.unpack_one("P"))
        self.add_field("FrameBufferBase", u.unpack_one("Q"))
        self.add_field("FrameBufferSize", u.unpack_one("P"))

    @property
    def Info(self):
        return GraphicsOutputModeInformation(self)

class GraphicsOutputProtocol(unpack.Struct):
    """Decode the EFI Graphics Output Protocol"""
    def __init__(self, ptr):
        super(GraphicsOutputProtocol, self).__init__()
        efi.unpack_pointer_table(self, ptr,
            'QueryMode', 'SetMode', 'Blt', 'ModePtr')

    @property
    def Mode(self):
        return GraphicsOutputProtocolMode(self.ModePtr)

def init():
    global gop
    uuid_mem = efi.alloc_var("16s", efi.EFI_GRAPHICS_OUTPUT_PROTOCOL_GUID.bytes_le)
    gop_ptr = efi.alloc_var("P", 0)
    efi.call(efi.system_table.BootServices.LocateProtocol, bits.memory_addr(uuid_mem), 0, bits.memory_addr(gop_ptr))
    gop = GraphicsOutputProtocol(struct.unpack("P", gop_ptr)[0])

def load(name="slides", bufsize=1024):
    """Load presentation slides"""
    global slides, current_slide, max_slide, saved_screen
    init()
    info = gop.Mode.Info
    saved_screen = bits.malloc(info.HorizontalResolution * info.VerticalResolution * 4)
    directory = os.path.join("/", name, "{}x{}".format(info.HorizontalResolution, info.VerticalResolution))
    slides = []
    for n in itertools.count(1):
        try:
            f = open(os.path.join(directory, str(n)), "rb")
            s = ''.join(iter(lambda:f.read(1024), ''))
        except IOError:
            max_slide = n - 2
            break
        m = bits.malloc(len(s))
        struct.pack_into("{}s".format(len(s)), m, 0, s)
        slides.append(m)
    current_slide = 0
    sizes = list(len(f) for f in slides)
    if set(sizes) != set([info.HorizontalResolution * info.VerticalResolution * 4]):
        print "Load error: Inconsistent buffer sizes = {}".format(sizes)
    readline.add_key_hook(bits.KEY_F10, resume)

def resume():
    """Start or resume a presentation"""
    global slides, current_slide, max_slide, saved_screen
    info = gop.Mode.Info
    efi.call(gop.Blt, gop._addr, bits.memory_addr(saved_screen), EfiBltVideoToBltBuffer, 0, 0, 0, 0, info.HorizontalResolution, info.VerticalResolution, 0)
    while True:
        if len(slides[current_slide]) != info.HorizontalResolution * info.VerticalResolution * 4:
            break
        efi.call(gop.Blt, gop._addr, bits.memory_addr(slides[current_slide]), EfiBltBufferToVideo, 0, 0, 0, 0, info.HorizontalResolution, info.VerticalResolution, 0)
        k = bits.get_key()
        if k == bits.KEY_ESC:
            break
        elif k in (bits.KEY_LEFT, bits.KEY_UP, bits.KEY_PAGE_UP):
            if current_slide > 0:
                current_slide -= 1
        elif k in (bits.KEY_RIGHT, bits.KEY_DOWN, bits.KEY_PAGE_DOWN, ord(' ')):
            if current_slide < max_slide:
                current_slide += 1
        elif k == bits.KEY_HOME:
            current_slide = 0
    efi.call(gop.Blt, gop._addr, bits.memory_addr(saved_screen), EfiBltBufferToVideo, 0, 0, 0, 0, info.HorizontalResolution, info.VerticalResolution, 0)
